
# Attempt at a simulated annealing ODE solver

# sidney42% R
# R version 2.7.1 (2008-06-23)

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

# load("cluster_cluster_model_n20_x3.RData") 
#
#     This is a 20 by 20 ODE coeff matrix named "model_x3_matrix"

# > load("cluster_cluster_model_n20_x3.RData")
# > ls()
# [1] "model_x3_matrix"
# > 
# 
# source("xoa_simulated_annealing_for_ODE_optimization_R_prog.R")
#
# final_matrix = runOpt()
# final_matrix


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

# bash-3.2$ r
# R version 2.7.1 (2008-06-23)
# Copyright (C) 2008 The R Foundation for Statistical Computing
# ISBN 3-900051-07-0
# 
# R is free software and comes with ABSOLUTELY NO WARRANTY.
# Type 'q()' to quit R.
# [Previously saved workspace restored]
#
# > ls()
# [1] "model_x3_matrix"
#
# > load("cluster_cluster_model_n20_x3.RData")
# > ls()
# [1] "model_x3_matrix"
#
# > source("xoa_simulated_annealing_for_ODE_optimization_R_prog.R")
# > runOpt()
#
# > ls()
# [1] "convert_NA_to_zero"                            
# [2] "eigenvalue_sum"                                
# [3] "eigenvalue_sum_of_real_parts"                  
# [4] "model_x3_matrix"                               
# [5] "runOpt"                                        
# [6] "scor"                                          
# [7] "simulated_annealing_optimization_of_ODE_system"
# > q()

# the optimization process from the paper.
# initmodel: a matrix with the coefficients of interactions between clusters (from cluster in rows, to cluster in columns)
#			 initmodel must have rownames and colnames that correspond to those in the rownames of the observed_values
# observed_values: a matrix with the clusters as rows and the observed data (time points) as columns
# init_values: a vector containing the initial state of all the clusters- in the order that the rows of initmodel are in
# eval.mode: the evaluation mode for performance. Currently supported are "correlation" (default) and "rmsd".
# tau: time step
# r1_nmodels: number of models to generate (optimization runs) for round 1
# r1_prob_reduction_factor: the amount the probability of retaining a deleterious mutation in round 1 decreases each step
# r1_SizeOfStep: a scaling factor for the random number generation for pertubation of parameters
# r1_init_prob: the initial probability for round 1
# r1_step_mod_mode: if perturbations are "relative" (default) then the amount of pertubation is 
#					random scaled relative to the current value
# r1_Nsteps: the number of steps taken for the optimization process
# r1_edge_removal_prob: the probability of removing an existing edge from the model at each step
# r1_edge_addition_prob: the probability of adding an edge to the model at each step


optimization_process <- function(initmodel, observed_values, init_values, cluster_counts, times, observation_indices=NA, eval.mode="correlation", tau=0.5, r1_nmodels=25, r1_prob_reduction_factor=0.002, r1_SizeOfStep=0.6, r1_init_prob=0.8, r1_step_mod_mode="relative", r1_Nsteps=5000, r1_edge_removal_prob=0.01, r1_edge_addition_prob=0.001, r2_nmodels=25, r2_prob_reduction_factor=0.002, r2_SizeOfStep=0.6, r2_init_prob=0.1, r2_step_mod_mode="relative", r2_Nsteps=25000, r2_edge_removal_prob=0.01, r2_edge_addition_prob=0.001, verbose=TRUE) {
	
	# this runs the two stage process described in the paper
	# first round starts with an initial model, runs 5000 iterations and produces 25 models.
	# the initial model for the second round is the best performance from the first round
	# the second round runs 25000 iterations and produces 25 models
	round1models = sapply(1:r1_nmodels, function (i) list(optimize_ode_matrix(initmodel, observed_values=observed_values, init_values=init_values, eval.mode=eval.mode, tau=tau, prob_reduction_factor=r1_prob_reduction_factor, SizeOfStep=r1_SizeOfStep, init_prob=r1_init_prob, step_mod_mode=r1_step_mod_mode, Niter=1, Nsteps=r1_Nsteps, edge_removal_prob=r1_edge_removal_prob, edge_addition_prob=r1_edge_addition_prob, verbose=verbose, cluster_counts=cluster_counts, times=times, observation_indices=observation_indices)))
	round1performance = sapply(1:r1_nmodels, function (i) round1models[[i]][[2]][r1_Nsteps])
	bestmodel = round1models[[which(round1performance==max(round1performance))]][[1]]
	round2models = sapply(1:r2_nmodels, function (i) list(optimize_ode_matrix(bestmodel, observed_values=observed_values, init_values=init_values, eval.mode=eval.mode, tau=tau, prob_reduction_factor=r2_prob_reduction_factor, SizeOfStep=r2_SizeOfStep, init_prob=r2_init_prob, step_mod_mode=r2_step_mod_mode, Niter=1, Nsteps=r2_Nsteps, edge_removal_prob=r2_edge_removal_prob, edge_addition_prob=r2_edge_addition_prob, verbose=verbose, cluster_counts=cluster_counts, times=times, observation_indices=observation_indices)))
	round2performance = sapply(1:r2_nmodels, function (i) round2models[[i]][[2]][r2_Nsteps])
	return(list(round1=list(round1models, round1performance), round2=list(round2models, round2performance)))

}

optimization_round2 <- function(round1models, observed_values, init_values, cluster_counts, times, observation_indices, selection_mode="max", r2_nmodels=25, r2_prob_reduction_factor=0.002, r2_SizeOfStep=0.6, r2_init_prob=0.1, r2_step_mod_mode="relative", r2_Nsteps=25000, r2_edge_removal_prob=0.01, r2_edge_addition_prob=0.001, verbose=TRUE) {
	round1performance = sapply(1:length(round1models), function (i) round1models[[i]][[2]][(length(round1models[[i]][[2]]))])
	
	# cat(round1performance, "\n", length(round1models[[1]][[2]]), "\n")
	
	if (selection_mode == "max") {
		bestmodel = round1models[[which(round1performance==max(round1performance))]][[1]]
	} else if (selection_mode == "min") {
		bestmodel = round1models[[which(round1performance==min(round1performance))]][[1]]
		#cat(min(round1performance), which(round1performance==min(round1performance)), "\n")
	}
	
	round2models = sapply(1:r2_nmodels, function (i) list(optimize_ode_matrix(bestmodel, observed_values=observed_values, init_values=init_values, prob_reduction_factor=r2_prob_reduction_factor, SizeOfStep=r2_SizeOfStep, init_prob=r2_init_prob, step_mod_mode=r2_step_mod_mode, Niter=1, Nsteps=r2_Nsteps, edge_removal_prob=r2_edge_removal_prob, edge_addition_prob=r2_edge_addition_prob, verbose=verbose, cluster_counts=cluster_counts, times=times, observation_indices=observation_indices)))
	round2performance = sapply(1:r2_nmodels, function (i) round2models[[i]][[2]][r2_Nsteps])
	return(list(round1=list(round1models, round1performance), round2=list(round2models, round2performance)))
}

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

convert_NA_to_zero <- function(mat) {
  mat_with_zeros <- mat
  mat_with_zeros[is.na(mat_with_zeros)] <- 0
  return(mat_with_zeros)
}

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

# target function to minimise
# that counts the number of eigenvalues below zero

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

eigenvalue_sum <- function(test_matrix) {
  	ev = eigen(test_matrix)
  	eigenvalue_vector = ev$val
  	sum = 0
  	for (i in 1:20){
  	    sum = sum + eigenvalue_vector[i]
  	}
  	return(sum)
} 

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

eigenvalue_sum_of_real_parts <- function(test_matrix) {
	ev = eigen(test_matrix)
	eigenvalue_vector = ev$val
	sum = 0
	printItem = sprintf('i=0, eigen_value_sum_of_real_parts running total = %2.6f', sum)
	print(printItem)
	for (i in 1:length(eigenvalue_vector) ) {
	  sum = sum + Re(eigenvalue_vector[i])
	  printItem = sprintf('i=%d, eigen_value_sum_of_real_parts running total = %2.6f', i, sum)
	  print(printItem)
    }
	return(sum)
}


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

# criterion for marking improvement
num_of_eigenvalues_less_than_zero <- function(test_matrix){
	ev = eigen(test_matrix)
	eigenvalue_vector = ev$val
	num_less_than_zero = 0
	# printItem = sprintf('i=0, num of eigen values whose real parts are less than zero = %d', num_less_than_zero)
	# print(printItem)
	for (i in 1:length(eigenvalue_vector) ) {
		real_part = Re(eigenvalue_vector[i])
		if ( real_part < 0 ) {
			num_less_than_zero = num_less_than_zero + 1
  		}
  	# printItem = sprintf('i=%d, part = %2.8f', i, real_part)
  	# print(printItem)
  	# printItem = sprintf('i=%d, num of eigen values whose real parts are less than zero (running count) = %d', i, num_less_than_zero)
  	# print(printItem)
	}
	return(num_less_than_zero)
}

maxeval_criterion <- function(test_matrix, ...) {
	maxev = max(Re(eigen(test_matrix)$values))
	return(-log(abs(maxev)))
}

ode_performance_criterion <- function(test_matrix, observed_values=NA, init_values=NA, 
										cluster_counts=NA, times=NA, observation_indices=NA, eval.mode="correlation",
										tau=NA, ...) {
	result = evaluate_matrices_ode(list(test_matrix), observed=observed_values, 
				counts=cluster_counts, init=init_values, times=times, observation_indices=observation_indices,
				eval.mode=eval.mode, tau=tau)
	if (eval.mode == "rmsd") {
		result = 1/result	
	}							
	return(result)
}

ode_performance_criterion_dual <- function(test_matrix, observed_values_dual=NA, init_values_dual=NA, 
										cluster_counts_dual=NA, times_dual=NA, eval.mode.dual="correlation", ...) {
	# simply an alternate that allows passing of different data for ODE performance evaluation
	#      when using the dual optimization method below
	return(evaluate_matrices_ode(list(test_matrix), observed=observed_values_dual, 
				counts=cluster_counts_dual, init=init_values_dual, times=times_dual, eval.mode=eval.mode.dual))
}


optimize_ode_matrix <- function(input_matrix, fitness_function="ode_performance_criterion", Niter=5, 
								Nsteps=15000, SizeOfStep=0.75, step_mod_mode="absolute", init_prob=0.5,
								prob_reduction_factor=0.01, edge_removal_prob=0, edge_addition_prob=0, 
								print_interval=500, verbose=FALSE, 
								only_optimize_nonzero_values=TRUE, ...) {
	
	initial_matrix                       = input_matrix
	initial_matrix_with_zeroes = convert_NA_to_zero(initial_matrix)

	# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	#Simulated annealing sequence	
	if (verbose) {
		printItem = sprintf('simulated_annealing_optimization_of_ODE_system STARTED')
		print(printItem)
	}
	
	current_matrix = initial_matrix_with_zeroes
	test_matrix    = current_matrix

	# highest (best) number we can get back is 20 values (out of a total
	# of 20 eigenvalues) having a real part that is less than zero

	#current_target_number = num_of_eigenvalues_less_than_zero(current_matrix)
	#current_target_number = maxeval_criterion(current_matrix)
	
	current_target_number = do.call(fitness_function, list(current_matrix, ...))
	
	best_matrix = current_matrix
	best_target_number = current_target_number

	best_matrix_over_all_cycles = best_matrix
	best_target_number_all_cycles = best_target_number

	if (verbose) {
		printItem = sprintf('initial_target_number (of starting matrix) = %f', current_target_number)
		print(printItem)
	}

	for (t in 1:Niter) {
		if (verbose) {
			print("-") 
			print("-") 
			print("%%%%%%%%%%% NEW ITERATION IN OUTER LOOP") 
			print("-") 
			printItem = sprintf('Iteration = %d ',t)
			print(printItem)
			print("-")
		} 
		
		trajectory = c()
		
	  	prob_P = init_prob
	  	current_matrix = initial_matrix_with_zeroes
		test_matrix    = current_matrix

		# highest (best) number we can get back is 20 values (out of a total
		# of 20 eigenvalues) having a real part that is less than zero

		current_target_number = do.call(fitness_function, list(current_matrix, ...))
		
		best_matrix = current_matrix
		best_target_number = current_target_number

  		for (d in 1:Nsteps) { 

  			prob_P = prob_P * (1 - prob_reduction_factor);

			if (only_optimize_nonzero_values) {
    			# random change to a non-zero coefficient
    			nonzero_indices = which(current_matrix != 0)
    		} else {
    			nonzero_indices = 1:length(current_matrix)
    		}
    		
    		if (runif(1,0,1) < edge_addition_prob) {
    			# this is FUBAR- we choose an index but it doesn't correspond to the nonzero indices
    			
    		    zero_indices = 1:length(current_matrix)
    			zero_indices = (!zero_indices %in% nonzero_indices)
    			# this is the fix though
    			nonzero_indices = zero_indices
				randomly_chosen_index_for_nonzero_coeff <- sample(1:length(nonzero_indices), 1)
				
				# we'd have to randomly happen on a value that improves things (or is randomly kept)
    			coeff_to_vary <- rnorm(1,0,2)
    			#cat("Adding an edge now...\n")
    		} else {
    			randomly_chosen_index_for_nonzero_coeff <- sample(1:length(nonzero_indices), 1)
    			coeff_to_vary <- current_matrix[nonzero_indices[randomly_chosen_index_for_nonzero_coeff]]
    		}


			# edge removal- if this is accepted (i.e. it makes fitness better or is randomly kept during annealing) 
			#      this edge is gone for the remainder of the annealing process!
			if (runif(1,0,1) < edge_removal_prob) {
				new_coeff = 0	
			} else {
       			# JM: note that there's an issue with using a proportional step size: 
    			#     it causes the values to mainly grow smaller, since the step size is getting
    			#     smaller as the magnitude decreases.
    			# modified to make the SizeOfStep a threshold on a random number
				if (step_mod_mode == "relative") {
    				step_to_use = rnorm(1, 0, SizeOfStep);
    				new_coeff <- coeff_to_vary + (coeff_to_vary * step_to_use);
				} else {
					step_to_use = rnorm(1,0,1)
					new_coeff = coeff_to_vary + (SizeOfStep*step_to_use)
				}
			}

    		test_matrix <- current_matrix
    		test_matrix[nonzero_indices[randomly_chosen_index_for_nonzero_coeff]] = new_coeff;
    		#test_target_number = num_of_eigenvalues_less_than_zero(test_matrix)    
			#test_target_number = maxeval_criterion(test_matrix)
			test_target_number = do.call(fitness_function, list(test_matrix, ...))

			remainder = d %% print_interval;
			if (verbose) {
			if (remainder == 0) {
				print("-") 
				printItem = sprintf('Step %d in inner loop', d)
				print(printItem)
				printItem = sprintf('randomly_chosen_index_for_nonzero_coeff = %d', randomly_chosen_index_for_nonzero_coeff)
				print(printItem)
				#printItem = sprintf('row_of_coeff_to_vary = %d', row_of_coeff_to_vary)
				#print(printItem)
				#printItem = sprintf('col_of_coeff_to_vary = %d', col_of_coeff_to_vary)
				#print(printItem)
				printItem = sprintf('coeff_to_vary=%2.8f', coeff_to_vary)
				print(printItem)
				#printItem = sprintf('plus_or_minus=%d', plus_or_minus)
				#print(printItem)
				printItem = sprintf('size of step_to_use=%2.3f', SizeOfStep*step_to_use)
				print(printItem)
				printItem = sprintf('new_coeff=%2.8f', new_coeff)
				print(printItem)
				printItem = sprintf('current_target_number=%g', current_target_number)
				print(printItem)
				printItem = sprintf('test_target_number=%f', test_target_number)
				print(printItem)
			}
			}

		    if ( test_target_number > current_target_number ) {
			        current_matrix <- test_matrix 
			        current_target_number <- test_target_number
		    }

		    if (current_target_number > best_target_number) {
				best_matrix = current_matrix
		        best_target_number = current_target_number
		    }


			if (remainder == 0 ) {
			    if ( test_target_number > current_target_number ) {
			       printItem = sprintf('current_matrix REPLACED by test_matrix since > criterion satisfied')
			       print(printItem)
			       printItem = sprintf('new current_target_number=%f', current_target_number)
			       print(printItem)
		    	}  else {
			       printItem = sprintf('current_matrix stays the same since > criterion NOT satisfied')
			       print(printItem)
		    	}
		    if (current_target_number > best_target_number) {
			       printItem = sprintf('best_matrix REPLACED by new current matrix')
			       print(printItem)
			       printItem = sprintf('new best_target_number=%f', best_target_number)
			       print(printItem)
		    	} else {
			       printItem = sprintf('best_matrix stays the same')
			       print(printItem)
	    		}
			}

			# from Wikipedia simulated annealing article:
			# Therefore, in practice the ideal cooling rate cannot be determined
			# beforehand, and should be empirically adjusted for each problem.

			# we have reached the max possible
			#    if (current_target_number == 20)
			#	break()


			# If we did not find improve our results, that is test_target_number
			# is no improvement, we still check to see if we exceed should do a
			# jump based on a random probability selection compared to a fn of the
			# current cooling temp

			# The probability cutoff P of a random jump starts very high, at
			# 1.0. That is, if we fall BENEATH the cutoff, we can make the
			# jump. Hence, at the start of our set of iterations, we can almost
			# always make the jump when we randomly select a prob value between 0
			# and 1. However, the value of P gets lowered at each iteration,
			# making the chance of a random jump less likely at each iteration,
			# ending at zero on the final iteration.

			if ( test_target_number <= current_target_number ) {
			     randNum_between_zero_and_one = runif(1,0,1)
			     if ( randNum_between_zero_and_one < prob_P ) {
			        current_matrix <- test_matrix 
			        current_target_number <- test_target_number
		         	if (current_target_number > best_target_number) {
			   	     	best_matrix = current_matrix
		             	best_target_number = current_target_number
        		 	}
				}
			}

			trajectory = c(trajectory, current_target_number)
			if (verbose) {
			if (remainder == 0 ) {
  			if ( test_target_number <= current_target_number ) {
     			printItem = sprintf('     test_target_number <= current_target_number, so we try a random move')
     			print(printItem)
     			printItem = sprintf('     randNum_between_zero_and_one = runif(1,0,1) = %2.6f', randNum_between_zero_and_one)
     			print(printItem)
     			printItem = sprintf('     current value of prob_P = %2.6f', prob_P)
     			print(printItem)
     		if ( randNum_between_zero_and_one < prob_P ) {
         		printItem = sprintf('     randNum selected is less than prob_P, so we REPLACE the current_matrix')
         		printItem = sprintf('     new current_target_number=%f', current_target_number)
         		print(printItem)
         		printItem = sprintf('     new current_target_number=%f', current_target_number)
         		# Note: since at this point we have 
         		#   test_target_number <= (the now former) current_target_number
         		# and we only change the best_matrix when 
         		#   current_target_number > best_target_number
         		# we know that the best_matrix stays the same.
         		printItem = sprintf('     best_matrix stays the same')
         		print(printItem)
     		} else {
         		printItem = sprintf('     randNum selected is not less than prob_P, so we the current_matrix stays the same')
         		print(printItem)
     		}
  			}
		}
		}
   
   
	# end of steps loop
	}	


  	if (best_target_number > best_target_number_all_cycles) {
       best_matrix_over_all_cycles = best_matrix
       best_target_number_all_cycles = best_target_number
       if (verbose) {
       		printItem = sprintf('OUTER LOOP CONCLUSION - we have REPLACEMENT in this iteration')
       		print(printItem)
       		printItem = sprintf('new best_matrix_over_all_cycles has been changed to best_matrix')
       		print(printItem)
       		printItem = sprintf('new best_target_number_all_cycles = %f', best_target_number_all_cycles)
       		print(printItem)
       	}
   	} else if (verbose ){
       		printItem = sprintf('OUTER LOOP CONCLUSION - no change in this iteration - best_matrix_over_all_cycles stays the same')
       		print(printItem)   
   }

	if (verbose) {
		print("-") 
		print("-") 
		printItem = sprintf('current OUTER ITERATION LOOP cycle ENDED; starting new cycle')
		print(printItem)
	}
# # end of the iterations loop
}

if (verbose) {
	print("-") 
	print("-") 

	printItem = sprintf('FINAL best_target_number_all_cycles = %f', best_target_number_all_cycles)
	print(printItem)
	print("-") 
	print("-") 

	printItem = sprintf('simulated_annealing_optimization_of_ODE_system ENDED')
	print(printItem)

	print("-") 
	print("-") 
}
	# end of function simulated_annealing_optimization_of_ODE_system = function(ode_matrix_with_zeroes)

	return(list(matrix=best_matrix, trajectory=trajectory))
}

optimize_ode_matrix_dual <- function(input_matrix, 
									fitness_function_1="ode_performance_criterion", 
									fitness_function_2="maxeval_criterion", Niter=5, 
									Nsteps=15000, SizeOfStep=0.75, step_mod_mode="absolute", init_prob=0.5,
									prob_reduction_factor=0.01, edge_removal_prob=0, edge_addition_prob=0, 
									print_interval=500, verbose=FALSE, 
									only_optimize_nonzero_values=TRUE, 
									results.filename=NA, ...) {

	# This SA routine uses two fitness functions simultaneously for optimization
	#     The way this is handled is by keeping track of each fitness score independently.
	#     If the perturbation improves BOTH scores, then it is kept.
	#     If it improves only one score or neither score, then it is left to the normal
	#          SA process for random determination of retention.
	
	initial_matrix                       = input_matrix
	initial_matrix_with_zeroes = convert_NA_to_zero(initial_matrix)

	# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	#Simulated annealing sequence	
	if (verbose) {
		printItem = sprintf('simulated_annealing_optimization_of_ODE_system STARTED')
		print(printItem)
	}
	
	current_matrix = initial_matrix_with_zeroes
	test_matrix    = current_matrix

	# highest (best) number we can get back is 20 values (out of a total
	# of 20 eigenvalues) having a real part that is less than zero

	#current_target_number = num_of_eigenvalues_less_than_zero(current_matrix)
	#current_target_number = maxeval_criterion(current_matrix)
	
	current_target_number_1 = do.call(fitness_function_1, list(current_matrix, ...))
	current_target_number_2 = do.call(fitness_function_2, list(current_matrix, ...))
	
	best_matrix = current_matrix
	best_target_number_1 = current_target_number_1
	best_target_number_2 = current_target_number_2

	best_matrix_over_all_cycles = best_matrix
	best_target_number_all_cycles_1 = best_target_number_1
	best_target_number_all_cycles_2 = best_target_number_2

	if (verbose) {
		printItem = sprintf('initial_target_number (of starting matrix) = %f %f', current_target_number_1, current_target_number_2)
		print(printItem)
	}

	for (t in 1:Niter) {
		if (verbose) {
			print("-") 
			print("-") 
			print("%%%%%%%%%%% NEW ITERATION IN OUTER LOOP") 
			print("-") 
			printItem = sprintf('Iteration = %d ',t)
			print(printItem)
			print("-")
		} 
		
		trajectory_1 = c()
		trajectory_2 = c()
		
	  	prob_P = init_prob
	  	current_matrix = initial_matrix_with_zeroes
		test_matrix    = current_matrix

		# highest (best) number we can get back is 20 values (out of a total
		# of 20 eigenvalues) having a real part that is less than zero

		current_target_number_1 = do.call(fitness_function_1, list(current_matrix, ...))
		current_target_number_2 = do.call(fitness_function_2, list(current_matrix, ...))
		
		best_matrix = current_matrix
		best_target_number_1 = current_target_number_1
		best_target_number_2 = current_target_number_2

  		for (d in 1:Nsteps) { 

  			prob_P = prob_P * (1 - prob_reduction_factor);

			if (only_optimize_nonzero_values) {
    			# random change to a non-zero coefficient
    			nonzero_indices = which(current_matrix != 0)
    		} else {
    			nonzero_indices = 1:length(current_matrix)
    		}
    		
    		if (runif(1,0,1) < edge_addition_prob) {
    		    zero_indices = 1:length(current_matrix)
    			zero_indices = (!zero_indices %in% nonzero_indices)
				randomly_chosen_index_for_nonzero_coeff <- sample(1:length(zero_indices), 1)
				
				# we'd have to randomly happen on a value that improves things (or is randomly kept)
    			coeff_to_vary <- rnorm(1,0,2)
    		} else {
    			randomly_chosen_index_for_nonzero_coeff <- sample(1:length(nonzero_indices), 1)
    			coeff_to_vary <- current_matrix[nonzero_indices[randomly_chosen_index_for_nonzero_coeff]]
    		}


			# edge removal- if this is accepted (i.e. it makes fitness better or is randomly kept during annealing) 
			#      this edge is gone for the remainder of the annealing process!
			if (runif(1,0,1) < edge_removal_prob) {
				new_coeff = 0	
			} else {
       			# JM: note that there's an issue with using a proportional step size: 
    			#     it causes the values to mainly grow smaller, since the step size is getting
    			#     smaller as the magnitude decreases.
    			# modified to make the SizeOfStep a threshold on a random number
				if (step_mod_mode == "relative") {
    				step_to_use = rnorm(1, 0, SizeOfStep);
    				new_coeff <- coeff_to_vary + (coeff_to_vary * step_to_use);
				} else {
					step_to_use = rnorm(1,0,1)
					new_coeff = coeff_to_vary + (SizeOfStep*step_to_use)
				}
			}

    		test_matrix <- current_matrix
    		test_matrix[nonzero_indices[randomly_chosen_index_for_nonzero_coeff]] = new_coeff;
    		#test_target_number = num_of_eigenvalues_less_than_zero(test_matrix)    
			#test_target_number = maxeval_criterion(test_matrix)
			test_target_number_1 = do.call(fitness_function_1, list(test_matrix, ...))
			test_target_number_2 = do.call(fitness_function_2, list(test_matrix, ...))

			remainder = d %% print_interval;
			if (verbose) {
			if (remainder == 0) {
				print("-") 
				printItem = sprintf('Step %d in inner loop', d)
				print(printItem)
				printItem = sprintf('randomly_chosen_index_for_nonzero_coeff = %d', randomly_chosen_index_for_nonzero_coeff)
				print(printItem)
				#printItem = sprintf('row_of_coeff_to_vary = %d', row_of_coeff_to_vary)
				#print(printItem)
				#printItem = sprintf('col_of_coeff_to_vary = %d', col_of_coeff_to_vary)
				#print(printItem)
				printItem = sprintf('coeff_to_vary=%2.8f', coeff_to_vary)
				print(printItem)
				#printItem = sprintf('plus_or_minus=%d', plus_or_minus)
				#print(printItem)
				printItem = sprintf('size of step_to_use=%2.3f', SizeOfStep*step_to_use)
				print(printItem)
				printItem = sprintf('new_coeff=%2.8f', new_coeff)
				print(printItem)
				printItem = sprintf('current_target_number_1=%g', current_target_number_1)
				print(printItem)
				printItem = sprintf('test_target_number_1=%f', test_target_number_1)
				print(printItem)
				printItem = sprintf('current_target_number_2=%g', current_target_number_2)
				print(printItem)
				printItem = sprintf('test_target_number_2=%f', test_target_number_2)
				print(printItem)
			}
			}
					    
		    # if either of these is an NA we'll skip this step- this still may produce problematic
		    #    behavior
		    if (!is.na(test_target_number_1) &! is.na(test_target_number_2)) {
		    
		    delta_1 = test_target_number_1 - current_target_number_1
		    delta_2 = test_target_number_2 - current_target_number_2
		    
		    # not sure what's going on here
		    keep_perturbation = FALSE
		    if (!is.na(delta_1) &! is.na(delta_2)) {
		    	# require that BOTH scores improve to keep
		    	if (delta_1 >= 0 && delta_2 >= 0) {
		    		keep_perturbation = TRUE
		    	}
		    }
		    
		    if (keep_perturbation) {
				current_matrix <- test_matrix 
				current_target_number_1 <- test_target_number_1
				current_target_number_2 <- test_target_number_2
		    }

			delta_best_1 = test_target_number_1 - best_target_number_1
			delta_best_2 = test_target_number_2 - best_target_number_2
			
			is_best_model = FALSE
			if (!is.na(delta_best_1) &! is.na(delta_best_2)) {
				if (delta_best_1 >= 0 && delta_best_2 >= 0) {
					is_best_model = TRUE
				}
			}

		    if (is_best_model) {
				best_matrix = current_matrix
		        best_target_number_1 = current_target_number_1
		        best_target_number_2 = current_target_number_2
		    }

			if (remainder == 0 ) {
			    if (keep_perturbation) {
			       printItem = sprintf('current_matrix REPLACED by test_matrix since > criterion satisfied')
			       print(printItem)
			       printItem = sprintf('new current_target_number_1=%f', current_target_number_1)
			       print(printItem)
			       printItem = sprintf('new current_target_number_2=%f', current_target_number_2)
			       print(printItem)
		    	}  else {
			       printItem = sprintf('current_matrix stays the same since > criterion NOT satisfied')
			       print(printItem)
		    }
		    
		    if (is_best_model) {
			       printItem = sprintf('best_matrix REPLACED by new current matrix')
			       print(printItem)
			       printItem = sprintf('new best_target_number_1=%f', best_target_number_1)
			       print(printItem)
			       printItem = sprintf('new best_target_number_2=%f', best_target_number_2)
			       print(printItem)
		    	} else {
			       printItem = sprintf('best_matrix stays the same')
			       print(printItem)
	    		}
			}

			# from Wikipedia simulated annealing article:
			# Therefore, in practice the ideal cooling rate cannot be determined
			# beforehand, and should be empirically adjusted for each problem.

			# we have reached the max possible
			#    if (current_target_number == 20)
			#	break()


			# If we did not find improve our results, that is test_target_number
			# is no improvement, we still check to see if we exceed should do a
			# jump based on a random probability selection compared to a fn of the
			# current cooling temp

			# The probability cutoff P of a random jump starts very high, at
			# 1.0. That is, if we fall BENEATH the cutoff, we can make the
			# jump. Hence, at the start of our set of iterations, we can almost
			# always make the jump when we randomly select a prob value between 0
			# and 1. However, the value of P gets lowered at each iteration,
			# making the chance of a random jump less likely at each iteration,
			# ending at zero on the final iteration.

			if (!keep_perturbation) {
			     randNum_between_zero_and_one = runif(1,0,1)
			     if ( randNum_between_zero_and_one < prob_P ) {
			        current_matrix <- test_matrix 
			        current_target_number_1 <- test_target_number_1
			        current_target_number_2 <- test_target_number_2
		         	if (is_best_model) {
			   	     	best_matrix = current_matrix
		             	best_target_number_1 = current_target_number_1
		             	best_target_number_2 = current_target_number_2
        		 	}
				}
			}

			trajectory_1 = c(trajectory_1, current_target_number_1)
			trajectory_2 = c(trajectory_2, current_target_number_2)
			
			if (verbose) {
			if (remainder == 0 ) {
  			if (!keep_perturbation) {
     			printItem = sprintf('     test_target_number <= current_target_number, so we try a random move')
     			print(printItem)
     			printItem = sprintf('     randNum_between_zero_and_one = runif(1,0,1) = %2.6f', randNum_between_zero_and_one)
     			print(printItem)
     			printItem = sprintf('     current value of prob_P = %2.6f', prob_P)
     			print(printItem)
     		if ( randNum_between_zero_and_one < prob_P ) {
         		printItem = sprintf('     randNum selected is less than prob_P, so we REPLACE the current_matrix')
         		printItem = sprintf('     new current_target_number_1=%f', current_target_number_1)
         		print(printItem)
         		printItem = sprintf('     new current_target_number_2=%f', current_target_number_2)
         		print(printItem)
         		# Note: since at this point we have 
         		#   test_target_number <= (the now former) current_target_number
         		# and we only change the best_matrix when 
         		#   current_target_number > best_target_number
         		# we know that the best_matrix stays the same.
         		printItem = sprintf('     best_matrix stays the same')
         		print(printItem)
     		} else {
         		printItem = sprintf('     randNum selected is not less than prob_P, so we the current_matrix stays the same')
         		print(printItem)
     		}
  			}
  			}
		}
		}
   
   
	# end of steps loop
	}	

	delta_best_all_1 = best_target_number_1 - best_target_number_all_cycles_1
	delta_best_all_2 = best_target_number_2 - best_target_number_all_cycles_2
	
	is_best_over_all = FALSE
	
	if (delta_best_all_1 >= 0 && delta_best_all_2 >= 0) {
		is_best_over_all = TRUE
	}

  	if (is_best_over_all) {
       best_matrix_over_all_cycles = best_matrix
       best_target_number_all_cycles_1 = best_target_number_1
       best_target_number_all_cycles_2 = best_target_number_2
       
       if (verbose) {
       		printItem = sprintf('OUTER LOOP CONCLUSION - we have REPLACEMENT in this iteration')
       		print(printItem)
       		printItem = sprintf('new best_matrix_over_all_cycles has been changed to best_matrix')
       		print(printItem)
       		printItem = sprintf('new best_target_number_all_cycles_1 = %f', best_target_number_all_cycles_1)
       		print(printItem)
       		printItem = sprintf('new best_target_number_all_cycles_2 = %f', best_target_number_all_cycles_2)
       		print(printItem)
       	}
   	} else if (verbose ){
       		printItem = sprintf('OUTER LOOP CONCLUSION - no change in this iteration - best_matrix_over_all_cycles stays the same')
       		print(printItem)   
   }

	if (verbose) {
		print("-") 
		print("-") 
		printItem = sprintf('current OUTER ITERATION LOOP cycle ENDED; starting new cycle')
		print(printItem)
	}
# # end of the iterations loop
}

if (verbose) {
	print("-") 
	print("-") 

	printItem = sprintf('FINAL best_target_number_all_cycles_1 = %f', best_target_number_all_cycles_1)
	print(printItem)
	printItem = sprintf('FINAL best_target_number_all_cycles_2 = %f', best_target_number_all_cycles_2)
	print(printItem)
	print("-") 
	print("-") 

	printItem = sprintf('simulated_annealing_optimization_of_ODE_system ENDED')
	print(printItem)

	print("-") 
	print("-") 
}
	# end of function simulated_annealing_optimization_of_ODE_system = function(ode_matrix_with_zeroes)
	result = list(matrix=best_matrix_over_all_cycles, trajectory_1=trajectory_1, trajectory_2=trajectory_2,
	 			performance_1=best_target_number_all_cycles_1, performance_2=best_target_number_all_cycles_2)

	if (!is.na(results.filename)) {
		save(result, file=results.filename)	
	}

	return(result)
}

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%